from copy import deepcopy
from datetime import datetime
from zoneinfo import ZoneInfo

from dateutil._common import weekday
from dateutil.relativedelta import FR, MO, SA, SU, TH, TU, WE, relativedelta
from django.test import TestCase
from edc_facility.import_holidays import import_holidays
from edc_visit_schedule.schedule.visit_collection import VisitCollection
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_appointment.models import Appointment
from edc_appointment_app.visit_schedule import visit_schedule1

from ..helper import Helper


class TestApptDatetimes(TestCase):
    helper_cls = Helper

    """Note: visit schedule has appointments on days 0, 1, 2, 3, etc.
    Default facility accepts appointments any day of the week.
    """

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        site_visit_schedules._registry = {}

    def register_visit_schedule(self, facility_name=None):
        """Overwrite facility name on each visit and register
        the modified visit_schedule.
        """
        visit_schedule = deepcopy(visit_schedule1)
        for schedule_name, schedule in visit_schedule.schedules.items():
            visit_collection = VisitCollection()
            for k, v in schedule.visits.items():
                v.facility_name = facility_name
                visit_collection.update({k: v})
            schedule.visits = visit_collection
            visit_schedule.schedules[schedule_name] = schedule
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule)
        visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule.name)
        for schedule_name, schedule in visit_schedule.schedules.items():
            for k, v in schedule.visits.items():
                self.assertEqual(v.facility_name, facility_name)

    def get_appt_datetimes(self, base_appt_datetime=None, subject_identifier=None):
        self.assertIsNotNone(base_appt_datetime)
        now = base_appt_datetime.astimezone(ZoneInfo("UTC"))
        self.helper = self.helper_cls(subject_identifier=subject_identifier, now=now)
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=subject_identifier)
        return [obj.appt_datetime for obj in appointments]

    def test_appointments_creation_dates(self):
        """Assert does not skip any days regardless of
        base appointment day.

        default facility (in tests) accepts appointments any day of the week.
        """
        self.register_visit_schedule(facility_name="7-day-clinic")
        for i in range(0, 7, 7):
            subject_identifier = f"12345{i}"
            dte = datetime(2017, 1, 7) + relativedelta(days=i)
            now = dte.astimezone(ZoneInfo("UTC"))
            self.helper = self.helper_cls(subject_identifier=subject_identifier, now=now)
            self.helper.consent_and_put_on_schedule()
            appointments = Appointment.objects.filter(subject_identifier=subject_identifier)
            appt_datetimes = [obj.appt_datetime for obj in appointments]
            base_appt_datetime = appt_datetimes[0]
            for index, appt_datetime in enumerate(appt_datetimes):
                self.assertEqual(
                    base_appt_datetime + relativedelta(days=index * 7), appt_datetime
                )

    def test_appointments_creation_dates2(self):
        """Assert skips SA, SU."""
        self.register_visit_schedule(facility_name="5-day-clinic")
        base_appt_datetime = datetime(2017, 1, 7)  # noqa
        self.assertTrue(weekday(base_appt_datetime.weekday()), SA)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="123456"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), MO)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TH)

        base_appt_datetime = datetime(2017, 1, 8)
        self.assertTrue(weekday(base_appt_datetime.weekday()), SU)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="1234567"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), MO)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TH)

        base_appt_datetime = datetime(2017, 1, 9)
        self.assertTrue(weekday(base_appt_datetime.weekday()), MO)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="12345678"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), MO)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TH)

        base_appt_datetime = datetime(2017, 1, 10)
        self.assertTrue(weekday(base_appt_datetime.weekday()), TU)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="123456789"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), FR)

        base_appt_datetime = datetime(2017, 1, 11)
        self.assertTrue(weekday(base_appt_datetime.weekday()), WE)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="1234567890"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), FR)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), MO)

    def test_appointments_creation_dates3(self):
        """Assert skips FR, SA, SU, MO."""
        self.register_visit_schedule(facility_name="3-day-clinic")
        base_appt_datetime = datetime(2017, 1, 7)  # noqa
        self.assertTrue(weekday(base_appt_datetime.weekday()), SA)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="123456"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TU)

        base_appt_datetime = datetime(2017, 1, 8)
        self.assertTrue(weekday(base_appt_datetime.weekday()), SU)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="1234567"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TU)

        base_appt_datetime = datetime(2017, 1, 9)
        self.assertTrue(weekday(base_appt_datetime.weekday()), MO)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="12345678"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TU)

        base_appt_datetime = datetime(2017, 1, 10)
        self.assertTrue(weekday(base_appt_datetime.weekday()), TU)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="123456789"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), TU)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TU)

        base_appt_datetime = datetime(2017, 1, 11)
        self.assertTrue(weekday(base_appt_datetime.weekday()), WE)
        appt_datetimes = self.get_appt_datetimes(
            base_appt_datetime=base_appt_datetime, subject_identifier="1234567890"
        )
        self.assertTrue(weekday(appt_datetimes[0].weekday()), WE)
        self.assertTrue(weekday(appt_datetimes[1].weekday()), TH)
        self.assertTrue(weekday(appt_datetimes[2].weekday()), FR)
        self.assertTrue(weekday(appt_datetimes[3].weekday()), TU)



from dateutil.relativedelta import relativedelta
from edc_visit_schedule import Schedule, Visit, VisitSchedule

from edc_appointment_app.visit_schedule.crfs import crfs, crfs_missed

visit_schedule4 = VisitSchedule(
    name="visit_schedule4",
    offstudy_model="edc_appointment_app.subjectoffstudy",
    death_report_model="edc_appointment_app.deathreport",
)

schedule4 = Schedule(
    name="three_monthly_schedule",
    onschedule_model="edc_appointment_app.onschedulethree",
    offschedule_model="edc_appointment_app.offschedulethree",
    appointment_model="edc_appointment.appointment",
    consent_model="edc_appointment_app.subjectconsent",
)

visits = [
    Visit(
        code="1000",
        title="Baseline",
        timepoint=0,
        rbase=relativedelta(days=0),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=0),
        add_window_gap_to_lower=True,
        requisitions=None,
        crfs=crfs,
        crfs_missed=crfs_missed,
        facility_name="7-day-clinic",
    )
]
for index, visit_code in [(3, "1030"), (6, "1060"), (9, "1090"), (12, "1120")]:
    visits.append(
        Visit(
            code=visit_code,
            title=f"Month {index}",
            timepoint=index,
            rbase=relativedelta(months=index),
            rlower=relativedelta(days=15),
            rupper=relativedelta(days=15),
            add_window_gap_to_lower=True,
            max_window_gap_to_lower=None,
            requisitions=None,
            crfs=crfs,
            crfs_missed=crfs_missed,
            facility_name="7-day-clinic",
            allow_unscheduled=True,
        )
    )
for visit in visits:
    schedule4.add_visit(visit)

visit_schedule4.add_schedule(schedule4)



from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.core.management.color import color_style
from edc_visit_schedule.constants import DAY1
from tqdm import tqdm

from edc_appointment.skip_appointments import SkipAppointments

style = color_style()


class Command(BaseCommand):
    help = "Update skipped appointments"

    def add_arguments(self, parser):
        parser.add_argument(
            "--model",
            dest="model",
            action="store",
            nargs="?",
            help="CRF model in label_lower format",
        )
        parser.add_argument(
            "--related_visit",
            dest="related_visit",
            default=settings.SUBJECT_VISIT_MODEL,
            action="store",
            nargs="?",
            help="Related visit model label_lower with or without app_label",
        )

    def handle(self, *args, **options) -> None:
        try:
            related_visit_model_cls = django_apps.get_model(options["related_visit"])
        except LookupError as e:
            raise CommandError(e)
        model = options["model"]
        try:
            app_label = model.split(".")[0]
        except KeyError:
            app_label = options["subject_visit"].split(".")[0]
        except AttributeError as e:
            raise CommandError(f"{e}. Specify a CRF model in label_lower format")
        crf_model_cls = django_apps.get_model(f"{app_label}.{model.split('.')[1]}")
        related_visits = related_visit_model_cls.objects.filter(
            visit_code=DAY1, visit_code_sequence=0
        ).order_by("subject_identifier")
        total = related_visits.count()
        for related_visit in tqdm(related_visits, total=total):
            try:
                obj = crf_model_cls.objects.get(subject_visit=related_visit)
            except ObjectDoesNotExist:
                pass
            else:
                SkipAppointments(obj).skip_to_next()



from django import template
from django.apps import apps as django_apps
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from edc_constants.constants import CLINIC, HOME, TELEPHONE

register = template.Library()


class ContinuationAppointmentUrlError(Exception):
    pass


class ContinuationAppointmentAnchor(template.Node):
    """Returns a reverse url for a continuation appointment
    if the appointment does not already exist.
    """

    def __init__(self, appointment, dashboard_type, extra_url_context):
        self.unresolved_appointment = template.Variable(appointment)
        self.unresolved_dashboard_type = template.Variable(dashboard_type)
        self.unresolved_extra_url_context = template.Variable(extra_url_context)

    @property
    def appointment_model(self):
        return django_apps.get_app_config("edc_appointment").appointment_model

    def render(self, context):
        self.appointment = self.unresolved_appointment.resolve(context)
        self.dashboard_type = self.unresolved_dashboard_type.resolve(context)
        self.registered_subject = self.appointment.registered_subject
        self.extra_url_context = self.unresolved_extra_url_context

        if not self.extra_url_context:
            self.extra_url_context = ""

        # does a continuation appointment exist? instance will be instance+1
        visit_code_sequences = []
        for appointment in self.appointment_model.objects.filter(
            registered_subject=self.appointment.registered_subject
        ):
            visit_code_sequences.append(int(appointment.visit_code_sequence))
        if (int(self.appointment.visit_code_sequence) + 1) in visit_code_sequences:
            anchor = ""
        else:
            view = (
                f"admin:{self.appointment._meta.app_label}_"
                f"{self.appointment._meta.module_name}_add"
            )
            try:
                url = reverse(view)
            except NoReverseMatch as e:
                raise ContinuationAppointmentUrlError(
                    "ContinuationAppointmentUrl Tag: NoReverseMatch while "
                    "rendering reverse for {self.appointment._meta.module_name}. "
                    f"Is model registered in admin? Got {e}."
                )
            else:
                # TODO: resolve error when using extra_url_context...give back
                # variable name ???
                visit_code_sequence = str(int(self.appointment.visit_code_sequence) + 1)
                rev_url = (
                    f"{url}?next=dashboard_url&dashboard_type={self.dashboard_type}"
                    f"&subject_identifier={self.appointment.subject_identifier}"
                    f"&visit_definition={self.appointment.visit_definition.pk}"
                    f"&visit_code_sequence={visit_code_sequence}"
                )
                anchor = f'<A href="{rev_url}">continuation</A>'
        return anchor


@register.tag(name="continuation_appointment_anchor")
def continuation_appointment_anchor(parser, token):
    """Compilation function for renderer ContinuationAppointmentUrl."""
    try:
        _, appointment, dashboard_type, extra_url_context = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires exactly 3 arguments" % token.contents.split()[0]
        )
    return ContinuationAppointmentAnchor(appointment, dashboard_type, extra_url_context)


@register.filter(name="appt_type")
def appt_type(value):
    """Filters appointment.appt_type."""
    if value == CLINIC:
        retval = "Clin"
    elif value == TELEPHONE:
        retval = "Tele"
    elif value == HOME:
        retval = "Home"
    else:
        retval = None
    return retval



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.db import models

from ..constants import CANCELLED_APPT
from ..exceptions import AppointmentWindowError
from ..utils import raise_on_appt_datetime_not_in_window

if TYPE_CHECKING:
    from ..models import Appointment


class WindowPeriodModelMixin(models.Model):
    """A model mixin declared with the Appointment model to managed
    window period calculations for appt_datetime.
    """

    window_period_checks_enabled: bool = True

    def save(self: Any, *args, **kwargs) -> None:
        if not kwargs.get("update_fields", None):
            self.raise_on_appt_datetime_not_in_window()
        super().save(*args, **kwargs)

    def raise_on_appt_datetime_not_in_window(self: Appointment) -> None:
        if (
            self.id
            and self.appt_status != CANCELLED_APPT
            and self.appt_datetime
            and self.timepoint_datetime
        ):
            if not self.ignore_window_period:
                try:
                    raise_on_appt_datetime_not_in_window(self)
                except AppointmentWindowError as e:
                    msg = f"{e} Perhaps catch this in the form"
                    raise AppointmentWindowError(msg)

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.utils import IntegrityError
from django.utils.timezone import is_naive
from edc_facility.facility import Facility, FacilityError
from edc_sites.valid_site_for_subject_or_raise import valid_site_for_subject_or_raise
from edc_visit_schedule.utils import is_baseline

from ..constants import SCHEDULED_APPT
from ..utils import (
    get_appointment_type_model_cls,
    get_appt_reason_default,
    get_appt_type_default,
)

if TYPE_CHECKING:
    from ..models import Appointment, AppointmentType


class CreateAppointmentError(Exception):
    pass


class CreateAppointmentDateError(Exception):
    pass


class AppointmentCreatorError(Exception):
    pass


if TYPE_CHECKING:
    from edc_visit_schedule.visit import Visit


class AppointmentCreator:
    def __init__(
        self,
        timepoint_datetime: datetime = None,
        timepoint: Decimal | None = None,
        visit: Visit | None = None,  # from edc_visit_schedule
        visit_code_sequence: int | None = None,
        facility: Facility | None = None,
        appointment_model: str = None,
        taken_datetimes: list[datetime] | None = None,
        subject_identifier: str = None,
        visit_schedule_name: str = None,
        schedule_name: str = None,
        default_appt_type: str | None = None,
        default_appt_reason: str | None = None,
        appt_status: str | None = None,
        appt_reason: str | None = None,
        suggested_datetime: datetime | None = None,
        skip_baseline: bool | None = None,
        ignore_window_period: bool | None = None,
    ):
        self._appointment = None
        self._appointment_model_cls = None
        self._default_appt_type = default_appt_type
        self._default_appt_reason = default_appt_reason
        self.skip_baseline = skip_baseline
        self.subject_identifier = subject_identifier
        self.visit_schedule_name = visit_schedule_name
        self.schedule_name: str = schedule_name
        self.appt_status = appt_status
        self.appt_reason = appt_reason
        self.appointment_model = appointment_model
        # already taken appt_datetimes for this subject
        self.taken_datetimes = taken_datetimes or []
        self.visit = visit
        self.visit_code_sequence = visit_code_sequence or 0
        self.timepoint = timepoint or self.visit.timepoint
        self.ignore_window_period = ignore_window_period
        if not isinstance(self.timepoint, Decimal):
            self.timepoint = Decimal(str(self.timepoint))

        try:
            if is_naive(timepoint_datetime):
                raise ValueError(
                    f"Naive datetime not allowed. {repr(self)}. " f"Got {timepoint_datetime}"
                )
            else:
                self.timepoint_datetime = timepoint_datetime
        except AttributeError:
            raise AppointmentCreatorError(
                f"Expected 'timepoint_datetime'. Got None. {repr(self)}."
            )
        # suggested_datetime (defaults to timepoint_datetime)
        # If provided, the rules for window period/rdelta relative
        # to timepoint_datetime still apply.
        if suggested_datetime and is_naive(suggested_datetime):
            raise ValueError(
                f"Naive datetime not allowed. {repr(self)}. " f"Got {suggested_datetime}"
            )
        else:
            self.suggested_datetime = suggested_datetime or self.timepoint_datetime
        self.facility = facility or visit.facility
        if not self.facility:
            raise AppointmentCreatorError(f"facility_name not defined. See {repr(visit)}")
        self.get_appointment()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(subject_identifier={self.subject_identifier}, "
            f"visit_code={self.visit.code}.{self.visit_code_sequence}@{int(self.timepoint)})"
        )

    def __str__(self):
        return self.subject_identifier

    def get_appointment(self) -> Appointment:
        return self.appointment

    @property
    def appointment(self) -> Appointment:
        """Returns a newly created or updated appointment model instance."""
        if not self._appointment:
            try:
                self._appointment = self.appointment_model_cls.objects.get(**self.options)
            except ObjectDoesNotExist:
                self._appointment = self._create()
            else:
                self._appointment = self._update(appointment=self._appointment)
        return self._appointment

    @property
    def options(self) -> dict:
        """Returns default options to "get" an existing
        appointment model instance.
        """
        options = dict(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            visit_code=self.visit.code,
            visit_code_sequence=self.visit_code_sequence,
            timepoint=self.timepoint,
        )
        if self.appt_status:
            options.update(appt_status=self.appt_status)
        return options

    def _create(self) -> Appointment:
        """Returns a newly created appointment model instance."""
        valid_site_for_subject_or_raise(self.options.get("subject_identifier"))
        try:
            with transaction.atomic():
                appointment = self.appointment_model_cls.objects.create(
                    facility_name=self.facility.name,
                    timepoint_datetime=self.timepoint_datetime,
                    appt_datetime=self.appt_datetime,
                    appt_type=self.default_appt_type,
                    appt_reason=self.appt_reason or self.default_appt_reason,
                    ignore_window_period=self.ignore_window_period or False,
                    **self.options,
                )
        except IntegrityError as e:
            raise CreateAppointmentError(
                f"An 'IntegrityError' was raised while trying to "
                f"create an appointment for subject '{self.subject_identifier}'. "
                f"Got {e}. Appointment create options were {self.options}"
            )
        return appointment

    def _update(self, appointment=None) -> Appointment:
        """Returns an updated appointment model instance."""
        if is_baseline(instance=appointment) and self.skip_baseline:
            pass
        else:
            appointment.appt_datetime = self.appt_datetime
            appointment.timepoint_datetime = self.timepoint_datetime
            appointment.save()
            appointment.refresh_from_db()
        return appointment

    @property
    def appt_datetime(self) -> datetime:
        """Returns an available appointment datetime.

        Raises an CreateAppointmentDateError if none.

        Returns an unadjusted suggested datetime if this is an
        unscheduled appointment.
        """
        if self.visit_code_sequence == 0 or self.visit_code_sequence is None:
            try:
                arw = self.facility.available_arr(
                    suggested_datetime=self.suggested_datetime,
                    forward_delta=self.visit.rupper,
                    reverse_delta=self.visit.rlower,
                    taken_datetimes=self.taken_datetimes,
                )
            except FacilityError as e:
                raise CreateAppointmentDateError(
                    f"{e} Visit={repr(self.visit)}. "
                    f"Try setting 'best_effort_available_datetime=True' on facility."
                )
        else:
            return self.suggested_datetime
        return arw.datetime

    @property
    def appointment_model_cls(self) -> Appointment:
        """Returns the appointment model class."""
        return django_apps.get_model("edc_appointment.appointment")

    @property
    def default_appt_type(self) -> AppointmentType | None:
        """Returns an AppointmentType instance or None for
        the default appointment type, e.g. 'clinic'.
        """
        if not self._default_appt_type:
            try:
                self._default_appt_type = get_appointment_type_model_cls().objects.get(
                    name=get_appt_type_default()
                )
            except ObjectDoesNotExist:
                self._default_appt_type = None
        return self._default_appt_type

    @property
    def default_appt_reason(self) -> str:
        """Returns a string that is the default appointment reason
        type, e.g. 'scheduled'.
        """
        if not self._default_appt_reason:
            try:
                self._default_appt_reason = get_appt_reason_default()
            except AttributeError:
                self._default_appt_reason = SCHEDULED_APPT
        return self._default_appt_reason



import re
from unittest.mock import patch

from django.contrib.auth.models import User
from django.urls import reverse
from django_webtest import WebTest
from edc_auth.auth_updater.group_updater import GroupUpdater, PermissionsCodenameError
from edc_facility import import_holidays
from edc_protocol import Protocol
from edc_reference import site_reference_configs
from edc_test_utils.webtest import login
from edc_visit_schedule import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.utils import (
    get_related_visit_model,
    get_related_visit_model_cls,
)

from edc_appointment.admin import AppointmentAdmin
from edc_appointment.auth_objects import codenames
from edc_appointment.constants import NEW_APPT
from edc_appointment.tests.helper import Helper
from edc_appointment.utils import get_appointment_model_cls
from edc_appointment_app.visit_schedule import visit_schedule1


def get_url_name():
    return "subject_dashboard_url"


class TestAdmin(WebTest):
    helper_cls = Helper
    extra_environ = {"HTTP_ACCEPT_LANGUAGE": "en"}

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self) -> None:
        super().setUp()
        self.user = User.objects.create_superuser("user_login", "u@example.com", "pass")
        self.user.is_active = True
        self.user.is_staff = True
        self.user.save()
        self.user.refresh_from_db()
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=Protocol().study_open_datetime,
        )
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": get_related_visit_model()}
        )

    def get_app_form(self, url_name=None, response=None):
        form = None
        if not response:
            response = self.app.get(
                reverse(url_name), extra_environ={"HTTP_ACCEPT_LANGUAGE": "en"}
            ).maybe_follow()
        for index, form in response.forms.items():
            if form.action == "/i18n/setlang/":
                continue
            else:
                break
        return form

    def login(self):
        form = self.get_app_form("admin:index")
        form["username"] = self.user.username
        form["password"] = "pass"  # nosec B105
        return form.submit()

    @patch.object(
        AppointmentAdmin,
        "get_subject_dashboard_url_name",
        side_effect=get_url_name,
    )
    def test_admin_ok(self, mock_get_subject_dashboard_url_name):
        subject_consent = self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.all()
        # there are 4 appts
        self.assertEqual(appointments.count(), 4)
        # all appts are new
        self.assertEqual(
            [appt.appt_status for appt in appointments],
            [NEW_APPT, NEW_APPT, NEW_APPT, NEW_APPT],
        )

        # login
        changelist_url_name = "edc_appointment_admin:edc_appointment_appointment_changelist"
        login(self, user=self.user, redirect_url=changelist_url_name)

        # go to changelist
        url = reverse(changelist_url_name)

        response = self.app.get(url, user=self.user)
        self.assertIn("1000", response.text)
        self.assertIn("2000", response.text)
        self.assertIn("3000", response.text)
        self.assertIn("4000", response.text)
        results = re.findall(r"Not started", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 5)  # one extra for listfilter
        results = re.findall(r"On time", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 1)  # one extra for listfilter
        results = re.findall(r"In Progress", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 1)  # one extra for listfilter

        get_related_visit_model_cls().objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        response = self.app.get(url, user=self.user)
        results = re.findall(r"Not started", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 4)  # one extra for listfilter
        results = re.findall(r"On time", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 2)  # one extra for listfilter
        results = re.findall(r"In Progress", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 2)  # one extra for listfilter

        url = f"{url}?q={subject_consent.subject_identifier}"
        response = self.app.get(url, user=self.user)
        results = re.findall(r"Not started", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 4)  # one extra for listfilter
        results = re.findall(r"On time", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 2)  # one extra for listfilter
        results = re.findall(r"In Progress", response.text, re.IGNORECASE)
        self.assertEqual(len(results), 2)  # one extra for listfilter

    def test_auth(self):
        group_updater = GroupUpdater(groups={})
        for codename in codenames:
            try:
                group_updater.get_from_dotted_codename(codename)
            except PermissionsCodenameError as e:
                self.fail(f"PermissionsCodenameError raised unexpectedly. Got {e}")



APPOINTMENT = "APPOINTMENT"
APPOINTMENT_VIEW = "APPOINTMENT_VIEW"
APPOINTMENT_EXPORT = "APPOINTMENT_EXPORT"

codenames = [
    "edc_appointment.add_appointment",
    "edc_appointment.change_appointment",
    "edc_appointment.view_appointment",
    "edc_appointment.view_historicalappointment",
]



from edc_constants.constants import CLINIC, HOME, HOSPITAL, NOT_APPLICABLE, TELEPHONE

list_data = {
    "edc_appointment.appointmenttype": [
        (CLINIC, "In clinic"),
        (HOME, "At home"),
        (HOSPITAL, "In hospital"),
        (TELEPHONE, "By telephone"),
        (NOT_APPLICABLE, "Not applicable"),
    ],
}



from importlib import import_module

from django.test import TestCase, override_settings
from edc_auth.auth_updater import AuthUpdater
from edc_auth.site_auths import site_auths


class TestAuths(TestCase):
    @override_settings(
        EDC_AUTH_SKIP_SITE_AUTHS=True,
        EDC_AUTH_SKIP_AUTH_UPDATER=True,
    )
    def test_load(self):
        site_auths.initialize()
        import_module("edc_appointment.auths")
        AuthUpdater(verbose=True)



from edc_list_data.model_mixins import ListModelMixin


class AppointmentType(ListModelMixin):
    class Meta:
        verbose_name = "Appointment Type"
        verbose_name_plural = "Appointment Types"



from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import post_delete, post_save, pre_delete
from django.dispatch import receiver
from edc_utils import formatted_datetime, get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.utils import get_related_visit_model_cls

from ..appointment_status_updater import (
    AppointmentStatusUpdater,
    AppointmentStatusUpdaterError,
)
from ..constants import IN_PROGRESS_APPT
from ..managers import AppointmentDeleteError
from ..skip_appointments import SkipAppointments
from ..utils import (
    InvalidModelForSkippedAppts,
    cancelled_appointment,
    get_allow_skipped_appt_using,
    get_appointment_model_cls,
    get_appointment_model_name,
    missed_appointment,
    update_unscheduled_appointment_sequence,
)


@receiver(post_save, weak=False, dispatch_uid="appointment_post_save")
def appointment_post_save(sender, instance, raw, update_fields, **kwargs):
    if not raw and isinstance(instance, get_appointment_model_cls()):
        if not update_fields:
            # use the AppointmentStatusUpdater to set all
            # other appointments to be NOT in progress
            try:
                if instance.appt_status == IN_PROGRESS_APPT:
                    AppointmentStatusUpdater(instance, clear_others_in_progress=True)
            except AttributeError as e:
                if "appt_status" not in str(e):
                    raise
            except AppointmentStatusUpdaterError:
                pass
        # try to create_missed_visit_from_appointment
        #  if appt_status == missed
        missed_appointment(instance)
        # try to delete subject visit
        # if appt status = CANCELLED_APPT
        cancelled_appointment(instance)


@receiver(post_save, weak=False, dispatch_uid="create_appointments_on_post_save")
def create_appointments_on_post_save(sender, instance, raw, created, using, **kwargs):
    """Method `Model.create_appointments` is not typically used.

    See schedule.put_on_schedule() in edc_visit_schedule.
    """
    if not raw and not kwargs.get("update_fields"):
        try:
            instance.create_appointments()
        except AttributeError as e:
            if "create_appointments" not in str(e):
                raise


@receiver(post_save, weak=False, dispatch_uid="update_appt_status_on_subject_visit_post_save")
def update_appt_status_on_related_visit_post_save(
    sender, instance, raw, update_fields, **kwargs
):
    if not raw and not update_fields:
        if isinstance(instance, (get_related_visit_model_cls(),)):
            try:
                AppointmentStatusUpdater(
                    instance.appointment,
                    change_to_in_progress=True,
                    clear_others_in_progress=True,
                )
            except AttributeError as e:
                if "appointment" not in str(e):
                    raise
            except AppointmentStatusUpdaterError:
                pass


@receiver(pre_delete, weak=False, dispatch_uid="appointments_on_pre_delete")
def appointments_on_pre_delete(sender, instance, using, **kwargs):
    if isinstance(instance, (get_appointment_model_cls(),)):
        if instance.visit_code_sequence == 0:
            schedule = site_visit_schedules.get_visit_schedule(
                instance.visit_schedule_name
            ).schedules.get(instance.schedule_name)
            onschedule_datetime = schedule.onschedule_model_cls.objects.get(
                subject_identifier=instance.subject_identifier
            ).onschedule_datetime
            try:
                offschedule_datetime = schedule.offschedule_model_cls.objects.get(
                    subject_identifier=instance.subject_identifier
                ).offschedule_datetime
            except ObjectDoesNotExist:
                raise AppointmentDeleteError(
                    f"Appointment may not be deleted. "
                    f"Subject {instance.subject_identifier} is on schedule "
                    f"'{instance.visit_schedule.verbose_name}.{instance.schedule_name}' "
                    f"as of '{formatted_datetime(onschedule_datetime)}'. "
                    f"Got appointment {instance.visit_code}.{instance.visit_code_sequence} "
                    f"datetime {formatted_datetime(instance.appt_datetime)}. "
                    f"Perhaps complete off schedule model "
                    f"'{instance.schedule.offschedule_model_cls().verbose_name.title()}' "
                    f"first."
                )
            else:
                if onschedule_datetime <= instance.appt_datetime <= offschedule_datetime:
                    raise AppointmentDeleteError(
                        f"Appointment may not be deleted. "
                        f"Subject {instance.subject_identifier} is on schedule "
                        f"'{instance.visit_schedule.verbose_name}.{instance.schedule_name}' "
                        f"as of '{formatted_datetime(onschedule_datetime)}' "
                        f"until '{formatted_datetime(get_utcnow())}'. "
                        f"Got appointment datetime "
                        f"{formatted_datetime(instance.appt_datetime)}. "
                    )
        else:
            # TODO: any conditions for unscheduled?
            pass


@receiver(post_delete, weak=False, dispatch_uid="appointments_on_post_delete")
def appointments_on_post_delete(sender, instance, using, **kwargs):
    if isinstance(instance, (get_appointment_model_cls(),)):
        if (
            not kwargs.get("update_fields")
            and sender._meta.label_lower == get_appointment_model_name()
        ):
            update_unscheduled_appointment_sequence(
                subject_identifier=instance.subject_identifier
            )


@receiver(
    post_save,
    weak=False,
    dispatch_uid="update_appointments_to_next_on_post_save",
)
def update_appointments_to_next_on_post_save(sender, instance, raw, created, using, **kwargs):
    if not raw and not kwargs.get("update_fields") and get_allow_skipped_appt_using():
        try:
            if get_allow_skipped_appt_using(instance):
                SkipAppointments(instance).skip_to_next()
        except InvalidModelForSkippedAppts:
            pass


@receiver(post_delete, weak=False, dispatch_uid="update_appointments_to_next_on_post_delete")
def update_appointments_to_next_on_post_delete(sender, instance, using, **kwargs):
    if get_allow_skipped_appt_using():
        try:
            if get_allow_skipped_appt_using(instance):
                SkipAppointments(instance).reset_skipped()
        except InvalidModelForSkippedAppts:
            pass



from .appointment import Appointment
from .appointment_type import AppointmentType
from .signals import (
    appointment_post_save,
    appointments_on_post_delete,
    appointments_on_pre_delete,
    create_appointments_on_post_save,
    update_appointments_to_next_on_post_save,
    update_appt_status_on_related_visit_post_save,
)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.conf import settings
from edc_utils import convert_php_dateformat, to_utc
from edc_utils.date import ceil_datetime, floor_secs
from edc_visit_schedule.schedule.window import (
    ScheduledVisitWindowError,
    UnScheduledVisitWindowError,
)
from edc_visit_schedule.utils import get_lower_datetime, get_upper_datetime

from ..constants import COMPLETE_APPT, INCOMPLETE_APPT

if TYPE_CHECKING:
    from edc_appointment.models import Appointment

UNSCHEDULED_WINDOW_ERROR = "unscheduled_window_error"
SCHEDULED_WINDOW_ERROR = "scheduled_window_error"


class WindowPeriodFormValidatorMixin:
    def validate_appt_datetime_in_window_period(self, appointment: Appointment, *args) -> None:
        self.datetime_in_window_or_raise(appointment, *args)

    @staticmethod
    def ignore_window_period_for_unscheduled(
        appointment: Appointment, proposed_appt_datetime: datetime
    ) -> bool:
        """Returns True if this is an unscheduled appt"""
        value = False
        if (
            appointment
            and appointment.visit_code_sequence > 0
            and appointment.next
            and appointment.next.appt_status in [INCOMPLETE_APPT, COMPLETE_APPT]
            and to_utc(proposed_appt_datetime) < appointment.next.appt_datetime
        ):
            value = True
        return value

    def datetime_in_window_or_raise(
        self,
        appointment: Appointment,
        proposed_appt_datetime: datetime,
        form_field: str,
    ):
        if proposed_appt_datetime:
            datetimestring = convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            try:
                appointment.schedule.datetime_in_window(
                    timepoint_datetime=appointment.timepoint_datetime,
                    dt=proposed_appt_datetime,
                    visit_code=appointment.visit_code,
                    visit_code_sequence=appointment.visit_code_sequence,
                    baseline_timepoint_datetime=self.baseline_timepoint_datetime(appointment),
                )
            except UnScheduledVisitWindowError as e:
                if not self.ignore_window_period_for_unscheduled(
                    appointment, proposed_appt_datetime
                ):
                    # TODO: fix the dates on this message to match e.message
                    lower = floor_secs(get_lower_datetime(appointment)).strftime(
                        datetimestring
                    )
                    upper = floor_secs(get_lower_datetime(appointment.next)).strftime(
                        datetimestring
                    )
                    self.raise_validation_error(
                        {
                            form_field: (
                                "Invalid. Expected a date/time between "
                                f"{lower} and {upper} (U). Got {e}."
                            )
                        },
                        UNSCHEDULED_WINDOW_ERROR,
                    )
            except ScheduledVisitWindowError:
                lower = floor_secs(get_lower_datetime(appointment)).strftime(datetimestring)
                upper = floor_secs(ceil_datetime(get_upper_datetime(appointment))).strftime(
                    datetimestring
                )
                proposed = floor_secs(proposed_appt_datetime).strftime(datetimestring)
                # TODO: check this is correctly limiting (e.g. requistions)
                self.raise_validation_error(
                    {
                        form_field: (
                            f"Invalid. Expected a date/time between {lower} and {upper} (S). "
                            f"Got {proposed}."
                        )
                    },
                    SCHEDULED_WINDOW_ERROR,
                )

    @staticmethod
    def baseline_timepoint_datetime(appointment: Appointment) -> datetime:
        return appointment.__class__.objects.first_appointment(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
        ).timepoint_datetime



# Generated by Django 2.0rc1 on 2017-11-28 19:53

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0011_auto_20171128_1054")]

    operations = [
        migrations.AlterModelOptions(
            name="appointment",
            options={"ordering": ("timepoint", "visit_code_sequence")},
        )
    ]



from edc_constants.constants import NOT_APPLICABLE

from .constants import (
    CANCELLED_APPT,
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    NEW_APPT,
    ONTIME_APPT,
    SCHEDULED_APPT,
    SKIPPED_APPT,
    UNSCHEDULED_APPT,
)

DEFAULT_APPT_REASON_CHOICES = (
    (SCHEDULED_APPT, "Scheduled (study-defined)"),
    (UNSCHEDULED_APPT, "Unscheduled / Routine"),
)

APPT_STATUS = (
    (NEW_APPT, "Not started"),
    (IN_PROGRESS_APPT, "In Progress"),
    (INCOMPLETE_APPT, "Incomplete"),
    (COMPLETE_APPT, "Done"),
    (CANCELLED_APPT, "Cancelled"),
    (SKIPPED_APPT, "Skipped as per protocol"),
)

APPT_TIMING = (
    (ONTIME_APPT, "On time (within window period)"),
    (MISSED_APPT, "Missed"),
    (NOT_APPLICABLE, "Not applicable"),
)

INFO_PROVIDER = (("subject", "Subject"), ("other", "Other person"))



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING, Any
from zoneinfo import ZoneInfo

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError
from edc_constants.constants import NOT_APPLICABLE

from .constants import MISSED_APPT, NEW_APPT, SKIPPED_APPT
from .models import Appointment, AppointmentType
from .utils import (
    InvalidModelForSkippedAppts,
    get_allow_skipped_appt_using,
    reset_appointment,
)

if TYPE_CHECKING:
    from edc_visit_schedule.models import VisitSchedule


class AnyCRF(Any):
    pass


class SkipAppointmentsError(Exception):
    pass


class SkipAppointments:
    """Set interim appointments as SKIPPED_APPT based on a date
    provided in a CRF model.
    """

    def __init__(self, crf_obj: AnyCRF):
        try:
            self.model, self.field_name = get_allow_skipped_appt_using(crf_obj)
        except InvalidModelForSkippedAppts:
            raise SkipAppointmentsError(
                "Appointments may not be skipped. "
                "settings.EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING="
                f"`{get_allow_skipped_appt_using()}`"
                f"Got model `{crf_obj._meta.label_lower}`."
            )
        self._last_crf_obj = None
        self._next_visit_code: str | None = None
        self.crf_obj = crf_obj
        self.model_cls = django_apps.get_model(self.model)
        self.related_visit_model_attr: str = self.crf_obj.related_visit_model_attr()
        self.appointment: Appointment = getattr(
            self.crf_obj, self.related_visit_model_attr
        ).appointment
        self.subject_identifier: str = self.appointment.subject_identifier
        self.visit_schedule_name: str = self.appointment.visit_schedule_name
        self.schedule_name: str = self.appointment.schedule_name
        self.next_appt_date: date = getattr(self.last_crf_obj, self.field_name)
        self.next_appt_datetime = datetime(
            year=self.next_appt_date.year,
            month=self.next_appt_date.month,
            day=self.next_appt_date.day,
            hour=8,
            minute=0,
            tzinfo=ZoneInfo("UTC"),
        )

    def skip_to_next(self):
        """Resets appointments and sets any as skipped until the
        date provided from the CRF.
        """
        if self.next_visit_code:
            self.reset_skipped()
            self.update_to_next()

    def reset_skipped(self):
        """Reset any Appointments previously where `appt_status`
        is SKIPPED_APPT.

        Also reset `appt_datetime` on any new appts (NEW_APPT).
        """

        # reset auto-created MISSED_APPT (appt_type__isnull=True)
        # TODO: this for-loop block may be removed once all auto-created
        #    missed appts are removed.
        for appointment in Appointment.objects.filter(
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            appt_type__isnull=True,
            appt_timing=MISSED_APPT,
            subject_identifier=self.subject_identifier,
        ).exclude(
            appt_status__in=[SKIPPED_APPT, NEW_APPT],
        ):
            reset_appointment(appointment)

        # reset SKIPPED_APPT or NEW_APPT
        for appointment in Appointment.objects.filter(
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            appt_status__in=[SKIPPED_APPT, NEW_APPT],
            subject_identifier=self.subject_identifier,
        ):
            try:
                reset_appointment(appointment)
            except IntegrityError as e:
                print(e)

    def update_to_next(self):
        """Update Appointments, set `appt_status` = SKIPPED_APPT up
        to the date provided from the CRF.

        The `appt_datetime` for the last appointment in the sequence
        will set to the date provided from the CRF and left as a
        NEW_APPT.
        .
        """
        appointment = self.appointment.next_by_timepoint
        while appointment:
            if self.update_appointment_as_next(appointment):
                break
            else:
                self.update_appointment_as_skipped(appointment)
            appointment = appointment.next_by_timepoint

    @property
    def last_crf_obj(self):
        """Return the CRF instance with the max date"""
        if not self._last_crf_obj:
            self._last_crf_obj = (
                self.model_cls.objects.filter(**self.query_opts)
                .order_by(self.field_name)
                .last()
            )
        return self._last_crf_obj

    @property
    def next_visit_code(self) -> str:
        """Return the next appointment visit_code entered on the last
        CRF instance.
        """
        if not self._next_visit_code:
            try:
                visit_schedule_obj: VisitSchedule = self.last_crf_obj.visitschedule
            except (AttributeError, ObjectDoesNotExist):
                self._next_visit_code = None
            else:
                self._next_visit_code = visit_schedule_obj.visit_code
        return self._next_visit_code

    @property
    def query_opts(self) -> dict[str, str]:
        return {
            f"{self.related_visit_model_attr}__subject_identifier": self.subject_identifier,
            f"{self.related_visit_model_attr}__visit_schedule_name": self.visit_schedule_name,
            f"{self.related_visit_model_attr}__schedule_name": self.schedule_name,
        }

    def update_appointment_as_next(self, appointment: Appointment) -> bool:
        if self.is_next_appointment(appointment) and appointment.appt_status in [
            NEW_APPT,
            SKIPPED_APPT,
        ]:
            appointment.appt_status = NEW_APPT
            appointment.appt_datetime = self.next_appt_datetime
            appointment.comment = ""
            appointment.save(update_fields=["appt_status", "appt_datetime", "comment"])
            return True
        return self.is_next_appointment(appointment)

    def is_next_appointment(self, appointment: Appointment) -> bool:
        """Return True if this is the appointment reported as the
        subject's next appointment.

        Match on visit_code.
        """
        return (
            appointment.visit_code == self.next_visit_code
            and appointment.visit_code_sequence == 0
        )

    def update_appointment_as_skipped(self, appointment: Appointment) -> None:
        if appointment.appt_status in [NEW_APPT, SKIPPED_APPT]:
            appointment.appt_type = AppointmentType.objects.get(name=NOT_APPLICABLE)
            appointment.appt_status = SKIPPED_APPT
            appointment.appt_timing = NOT_APPLICABLE
            appointment.comment = (
                "based on date reported at " f"{self.last_crf_obj.related_visit.visit_code}"
            )
            appointment.save(
                update_fields=["appt_type", "appt_status", "appt_timing", "comment"]
            )
